import 'package:equations_solver/blocs/plot_zoom/plot_zoom.dart';
import 'package:equations_solver/blocs/polynomial_solver/polynomial_solver.dart';
import 'package:equations_solver/blocs/textfield_values/textfield_values.dart';
import 'package:equations_solver/localization/localization.dart';
import 'package:equations_solver/routes/polynomial_page/polynomial_input_field.dart';
import 'package:equations_solver/routes/utils/body_pages/equation_text_formatter.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

/// This widget contains a series of [PolynomialInputField] widgets needed to
/// parse the coefficients of the polynomial to be solved.
class PolynomialDataInput extends StatelessWidget {
  /// Creates a [PolynomialDataInput] widget.
  const PolynomialDataInput({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final polynomialType = context.read<PolynomialBloc>().polynomialType;

    switch (polynomialType) {
      case PolynomialType.linear:
        return const _InputWidget(
          inputLength: 2,
          equationTemplate: 'ax + b',
        );
      case PolynomialType.quadratic:
        return const _InputWidget(
          inputLength: 3,
          equationTemplate: 'ax^2 + bx + c',
        );
      case PolynomialType.cubic:
        return const _InputWidget(
          inputLength: 4,
          equationTemplate: 'ax^3 + bx^2 + cx + d',
        );
      case PolynomialType.quartic:
        return const _InputWidget(
          inputLength: 5,
          equationTemplate: 'ax^4 + bx^3 + cx^2 + dx + e',
        );
    }
  }
}

/// The actual input container.
class _InputWidget extends StatefulWidget {
  /// How many input fields the screen has to display.
  final int inputLength;

  /// The 'form' of the equation. This can be, for example, `ax^2 + bx + c` in
  /// the case of a quadratic polynomial.
  final String equationTemplate;

  /// Creates a [_InputWidget] widget.
  const _InputWidget({
    required this.inputLength,
    required this.equationTemplate,
  });

  @override
  __InputWidget createState() => __InputWidget();
}

class __InputWidget extends State<_InputWidget> {
  /// Controllers of the various input fields are "cached" because they'll never
  /// change during the lifetime of the widget.
  late final controllers = List<TextEditingController>.generate(
    widget.inputLength,
    _generateTextController,
    growable: false,
  );

  /// This is displayed at the top of the input box.
  late final cachedEquationTitle = Padding(
    padding: const EdgeInsets.only(bottom: 20),
    child: EquationTextFormatter(
      equation: widget.equationTemplate,
    ),
  );

  /// Form validation key.
  final formKey = GlobalKey<FormState>();

  /// The input fields are placed inside a [Wrap] widget which will take care of
  /// responsively laying out children in the UI according with the available
  /// width.
  ///
  /// This is cached because the number of input fields won't change.
  late final Widget cachedWrap = _InputFieldsWrapWidget(
    controllers: controllers,
    inputLength: widget.inputLength,
  );

  /// Generates the controllers and hooks them to the [TextFieldValuesCubit] in
  /// order to cache the user input.
  TextEditingController _generateTextController(int index) {
    // Initializing with the cached value, if any
    final controller = TextEditingController(
      text: context.read<TextFieldValuesCubit>().getValue(index),
    );

    // Listener that updates the value
    controller.addListener(() {
      context.read<TextFieldValuesCubit>().setValue(
            index: index,
            value: controller.text,
          );
    });

    return controller;
  }

  /// Validates the input and, if it's valid, sends the data to the bloc.
  void _processInput() {
    if (formKey.currentState?.validate() ?? false) {
      final event = PolynomialSolve(
        coefficients: controllers.map<String>((c) => c.text).toList(),
      );

      // Valid input
      context.read<PolynomialBloc>().add(event);
    } else {
      context.read<PolynomialBloc>().add(const PolynomialClean());

      // Error message
      ScaffoldMessenger.of(context).showSnackBar(
        SnackBar(
          content: Text(context.l10n.polynomial_error),
          duration: const Duration(seconds: 2),
        ),
      );
    }
  }

  /// Form cleanup.
  void _cleanInput() {
    for (final controller in controllers) {
      controller.clear();
    }

    formKey.currentState?.reset();
    context.read<PolynomialBloc>().add(const PolynomialClean());
    context.read<PlotZoomCubit>().reset();
    context.read<TextFieldValuesCubit>().reset();

    FocusScope.of(context).unfocus();
  }

  @override
  void dispose() {
    for (final controller in controllers) {
      controller.dispose();
    }

    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return BlocListener<TextFieldValuesCubit, Map<int, String>>(
      listener: (_, state) {
        if (state.isEmpty) {
          for (final controller in controllers) {
            controller.clear();
          }
        }
      },
      child: Column(
        mainAxisSize: MainAxisSize.min,
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          // Some space from the top
          const SizedBox(height: 40),

          // The title
          cachedEquationTitle,

          // Responsively placing input fields using 'Wrap'
          Padding(
            padding: const EdgeInsets.only(left: 60, right: 60),
            child: Form(
              key: formKey,
              child: cachedWrap,
            ),
          ),

          // A "spacer" widget
          const SizedBox(height: 40),

          // Two buttons needed to "solve" and "clear" the equation
          Row(
            mainAxisAlignment: MainAxisAlignment.center,
            children: [
              // Solving the polynomial
              ElevatedButton(
                key: const Key('Polynomial-button-solve'),
                onPressed: _processInput,
                child: Text(context.l10n.solve),
              ),

              // Some spacing
              const SizedBox(width: 30),

              // Cleaning the inputs
              ElevatedButton(
                key: const Key('Polynomial-button-clean'),
                onPressed: _cleanInput,
                child: Text(context.l10n.clean),
              ),
            ],
          ),
        ],
      ),
    );
  }
}

/// Generates the input fields of the coefficients and puts them in a [Wrap].
class _InputFieldsWrapWidget extends StatelessWidget {
  /// The [TextEditingController]s of each input field.
  final List<TextEditingController> controllers;

  /// How many input fields the screen has to display.
  final int inputLength;

  const _InputFieldsWrapWidget({
    Key? key,
    required this.controllers,
    required this.inputLength,
  }) : super(key: key);

  /// Increments by [index] the char code unit to get a specific letter. For
  /// example:
  ///
  ///   - [index] = 0 -> 'a'
  ///   - [index] = 1 -> 'b'
  ///   - [index] = 2 -> 'c'
  ///   - [index] = 6 -> 'g'
  String placeholderLetter(int index) {
    const firstLetter = 'a';

    return String.fromCharCode(firstLetter.codeUnitAt(0) + index);
  }

  @override
  Widget build(BuildContext context) {
    return Wrap(
      spacing: 30,
      alignment: WrapAlignment.center,
      children: [
        for (var i = 0; i < inputLength; ++i)
          PolynomialInputField(
            // This key is used for testing purposes
            key: Key('PolynomialInputField-coefficient-$i'),
            controller: controllers[i],
            placeholder: placeholderLetter(i),
          ),
      ],
    );
  }
}
